#include "bintree.h"

#include <iostream>

using namespace std;

namespace bst {

struct Tree_node
{
    string key;
    unsigned value;
    Tree left;
    Tree right;
};

bool member(Tree tree, const string& key)
{
    return find_subtree(tree, key) != nullptr;
}

unsigned lookup(Tree tree, const string& key)
{
    Tree node = find_subtree(tree, key);
    if (node == nullptr)
        throw Not_found(key);
    else return node->value;
}

void insert(Tree& tree, const string& key, unsigned value)
{
    Tree& node = find_subtree_mut(tree, key);
    if (node == nullptr) {
        node = make_shared<Tree_node>();
        node->key = key;
    }
    node->value = value;
}

Tree find_subtree(Tree tree, const string& key)
{
    while (tree != nullptr) {
        if (key < tree->key) tree = tree->left;
        else if (key > tree->key) tree = tree->right;
        else return tree;
    }

    return nullptr;
}

Tree& find_subtree_mut(Tree& tree, const string& key)
{
    if (tree == nullptr) return tree;
    if (key < tree->key) return find_subtree_mut(tree->left, key);
    if (key > tree->key) return find_subtree_mut(tree->right, key);
    return tree;
}

Tree& find_subtree_mut_iter(Tree& tree, const string& key)
{
    if (tree == nullptr || tree->key == key)
        return tree;

    Tree curr = tree;
    for (;;) {
        if (key < curr->key) {
            if (curr->left == nullptr || key == curr->left->key)
                return curr->left;
            else
                curr = curr->left;
        } else {
            if (curr->right == nullptr || key == curr->right->key)
                return curr->right;
            else
                curr = curr->right;
        }
    }
}

void get_keys_helper(Tree tree, vector<string>& out)
{
    if (tree != nullptr) {
        get_keys_helper(tree->left, out);
        out.push_back(tree->key);
        get_keys_helper(tree->right, out);
    }
}

vector<string> get_keys(Tree tree)
{
    vector<string> result;
    get_keys_helper(tree, result);
    return result;
}

}
