#pragma once

#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

namespace bst {

// Forward declaration of Tree_node lets us refer Tree_node in the definition
// of Tree before defining Tree_node.
struct Tree_node;

// A Tree will be represented as a shared_ptr to a Tree_node. Tree_node
// itself is defined in bintree.cpp, because clients don't need to see the
// definition.
using Tree = std::shared_ptr<Tree_node>;

// An exception thrown by lookup(Tree, const std::string&) when the key isn't
// found.
struct Not_found : std::runtime_error
{
    Not_found(const std::string& key) :
            runtime_error("key not found: " + key) { }
};

// Is the key present in the tree?
bool member(Tree, const std::string&);

// Returns the value associated with the key in the tree, or throws Not_found
// if the key is absent.
unsigned lookup(Tree, const std::string&);

// Inserts or updates the value associated with a key in the tree.
void insert(Tree&, const std::string&, unsigned);

// Finds the subtree associated with a particular key.
Tree find_subtree(Tree, const std::string& key);
Tree& find_subtree_mut(Tree&, const std::string& key);

// Returns all keys in order as a vector.
std::vector<std::string> get_keys(Tree);

}