#include "string_list.h"
#include <catch.h>

using namespace string_list;

TEST_CASE("is_empty recognizes empty vs non")
{
    List mt;
    CHECK(is_empty(mt));

    List lst = cons("a", mt);
    CHECK_FALSE(is_empty(lst));
}

TEST_CASE("is_cons recognizes empty vs non")
{
    List mt;
    CHECK_FALSE(is_cons(mt));

    List lst = cons("a", mt);
    CHECK(is_cons(lst));
}

TEST_CASE("cons builds list")
{
    List lst = cons("a", cons("b", nullptr));
    CHECK(is_cons(lst));
    CHECK("a" == lst->first);
    CHECK(is_cons(lst->rest));
    CHECK("b" == lst->rest->first);
    CHECK(is_empty(lst->rest->rest));
}

TEST_CASE("Length of 0 1 2 3 works")
{
    List lst;
    CHECK(0 == length(lst));
    lst = cons("c", lst);
    CHECK(1 == length(lst));
    lst = cons("b", lst);
    CHECK(2 == length(lst));
    lst = cons("a", lst);
    CHECK(3 == length(lst));
}

TEST_CASE("append(empty, empty) yields empty")
{
    List mt;
    List result = append(mt, mt);
    CHECK(is_empty(result));
}

TEST_CASE("Append(empty, non-empty) yields same")
{
    List mt;
    List lst1 = cons("a", cons("b", mt));
    List lst2 = append(mt, lst1);

    CHECK(is_cons(lst2));
    CHECK("a" == lst2->first);
    CHECK(is_cons(lst2->rest));
    CHECK("b" == lst2->rest->first);
    CHECK(is_empty(lst2->rest->rest));

    CHECK(lst1 == lst2);
}

TEST_CASE("Append(non-empty, empty) yields copy")
{
    List mt;
    List lst1 = cons("a", cons("b", mt));
    List lst2 = append(lst1, mt);

    CHECK(is_cons(lst2));
    CHECK("a" == lst2->first);
    CHECK(is_cons(lst2->rest));
    CHECK("b" == lst2->rest->first);
    CHECK(is_empty(lst2->rest->rest));

    CHECK(lst1 != lst2);
}

TEST_CASE("Append(non-empty, non-empty) yields half copy")
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2 = cons("c", cons("d", nullptr));
    List lst3 = append(lst1, lst2);

    CHECK(lst1 != lst3);

    CHECK(is_cons(lst3));
    CHECK("a" == lst3->first);
    CHECK(is_cons(lst3->rest));
    CHECK("b" == lst3->rest->first);

    CHECK(lst2 == lst3->rest->rest);
}

TEST_CASE("Concat(empty, empty) yields empty")
{
    List lst1;
    List lst2;
    concat(lst1, lst2);
    CHECK(is_empty(lst1));
    CHECK(is_empty(lst2));
}

TEST_CASE("Concat(empty, non-empty) yields alias")
{
    List lst1;
    List lst2 = cons("a", cons("b", nullptr));
    concat(lst1, lst2);

    CHECK(lst1 == lst2);
    CHECK(is_cons(lst1));
    CHECK(is_cons(lst1->rest));
    CHECK(is_empty(lst1->rest->rest));
}

TEST_CASE("Concat(non-empty, empty) no change")
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2;
    concat(lst1, lst2);

    CHECK(2 == length(lst1));
    CHECK(0 == length(lst2));
}

TEST_CASE("Concat(non-empty, non-empty) yields alias")
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2 = cons("c", cons("d", nullptr));

    concat(lst1, lst2);

    CHECK(4 == length(lst1));
    CHECK(2 == length(lst2));

    CHECK(is_cons(lst1->rest));
    CHECK(is_cons(lst1->rest->rest));
    CHECK(lst1->rest->rest == lst2);
}
